##
# @file		bshow_param.tcl
#
# @brief	Procedures to read and write parameter files
#
# @author	Bernard Heymann
# @date		Created: 20010516
# @date		Modified: 20190911

set paramfile ""
set paramdir [pwd]
set project_item ""
set fom_cutoff 0
set tilt_axis 0
set tilt_angle 0
set tilt_offset 0
set show_tilt_axis 1

## @brief Menu for opening and closing parameter files
#

proc setupParameterMenu {} {
	global theimg
	
	.menuBar add cascade -menu .menuBar.param -label "Micrograph" -underline 0
	menu .menuBar.param -tearoff 0
	.menuBar.param add command -label "Read parameters" \
		-command { openParam } \
		-underline 0 -accelerator "Ctrl-r"
	.menuBar.param add command -label "Append parameters" \
		-command { appendParam } \
		-underline 0 -accelerator "Ctrl-a"
	.menuBar.param add command -label "Write parameters" \
		-command { writeParam } \
		-underline 0 -accelerator "Ctrl-w"
	.menuBar.param add command -label "Edit parameters" \
		-command { editParam } \
		-underline 0 -accelerator "Ctrl-e"
	.menuBar.param add command -label "Change image" \
		-command { selectAFile } \
		-underline 0 -accelerator "Ctrl-a"
	.menuBar.param add command -label "Fit CTF" \
		-command { CTF $theimg } \
		-underline 0 -accelerator "Ctrl-c"
	.menuBar.param add command -label "Particles" \
		-command { Particles } \
		-underline 0 -accelerator "Ctrl-p"
	.menuBar.param add command -label "Filaments" \
		-command { Filaments } \
		-underline 0 -accelerator "Ctrl-f"
	.menuBar.param add command -label "Helix" \
		-command { Helix } \
		-underline 0 -accelerator "Ctrl-h"
	.menuBar.param add command -label "Crystallography" \
		-command { Crystallography } \
		-underline 0 -accelerator "Ctrl-x"
	.menuBar.param add command -label "Tomography" \
		-command { Tomography } \
		-underline 0 -accelerator "Ctrl-t"
}


## @brief Opens a micrograph parameter file for reading.
#

proc openParam { } {
	global paramfile paramdir
	set currdir $paramdir
	if { [string length $currdir] < 2 } { set currdir [pwd] }
    set types {
		{"STAR files"		{.star .STAR}	}
		{"XML files"		{.xml .XML}	}
		{"EMX files"		{.emx .EMX}	}
	}
	set paramfile [tk_getOpenFile -title "Open parameter file" -filetypes $types \
			 -initialdir $currdir -defaultextension .star]
	if { [string length $paramfile] < 1 } { return }
	loadParam $paramfile
	Update 0
}

## @brief Opens a micrograph parameter file for reading.
#
# @param	file			parameter file name.

proc loadParam { file } {
	global paramfile paramdir
	global project_item
	set project_item "all"
	set paramfile $file
	Bmg read $paramfile
	set paramdir [file dirname $paramfile]
	set paramfile [file tail $paramfile]
#	puts "$paramdir $paramfile"
	selectAFile
}

## @brief Appends one or more micrograph parameter files.
#

proc appendParam { } {
	global paramfile paramdir
	set currdir $paramdir
	if { [string length $currdir] < 2 } { set currdir [pwd] }
    set types {
		{"STAR files"		{.star .STAR}	}
		{"XML files"		{.xml .XML}	}
		{"EMX files"		{.emx .EMX}	}
	}
	set addfiles [tk_getOpenFile -title "Append parameter files" -filetypes $types \
			 -multiple true -initialdir $currdir -defaultextension .star]
	if { [string length $addfiles] < 1 } { return }
	puts $addfiles
#	loadParam $paramfile
	Bmg add $addfiles
	Update 0
}

proc tag_line { w } {
	set cur [$w index current]
	$w tag delete seline
	$w tag add seline "$cur linestart" "$cur lineend"
	$w tag configure seline -background yellow
}

## @brief Dialog box to select an image file.
#

proc selectAFile { } {
	global filename
	global imgtype
	global project_item
	if { [Bmg exists] } {
		set img_files [Bmg get all image_filenames]
	} elseif { [Bmodel exists] } {
		set img_files [Bmodel get all image_filenames]
	} else {
		return
	}
#	puts "$img_files"
	set oldfilename $filename
	if { [llength $img_files] < 1 } {
		puts "Error: No image files found!"
		if { [Bmodel exists] } {
			Bmodel set all map $filename
		}
	} elseif { [llength $img_files] < 2 } {
		set list [split $img_files ":"]
		set filename [lindex $list 0]
		set imgtype [lindex $list 1]
		set project_item "[lindex $list 2]:[lindex $list 3]"
#		puts "$filename ($imgtype) $project_item"
	} else {
		set w .wsel
		catch {destroy $w}
		toplevel $w
		wm title $w "Select an image"
		wm iconname $w "Select"

		label $w.msg -wraplength 4i -justify left \
			-text {Double click on the desired file name}
		pack $w.msg -side top -fill x -pady 2m
		
		frame $w.file
		scrollbar $w.file.xscroll -command "$w.file.list xview" -orient horizontal
		scrollbar $w.file.yscroll -command "$w.file.list yview"
		text $w.file.list -width 80 -height 20 -relief sunken -borderwidth 3 \
			-xscrollcommand "$w.file.xscroll set" \
			-yscrollcommand "$w.file.yscroll set" -wrap none
		pack $w.file.xscroll -side bottom -expand 0 -fill x
		pack $w.file.yscroll -side right -expand 0 -fill y
		pack $w.file.list -side left -expand 1 -fill both
		pack $w.file -side top -fill both -pady 2m -expand 1

		frame $w.buttons
		button $w.buttons.ok -text OK -command { setSelectionAndDestroy }
		button $w.buttons.cancel -text Cancel -command { destroy .wsel }
		pack $w.buttons -side bottom -fill x -pady 2m
		pack $w.buttons.ok $w.buttons.cancel -side left -expand 1
		
		foreach name $img_files {
			$w.file.list insert end "$name"
#			puts "$filename $name"
			if [string length $filename] {
				if [regexp $filename $name] {
					$w.file.list see end
					tag_line $w.file.list
#					puts $filename
				}
			}
			$w.file.list insert end "\n"
		}
		$w.file.list delete [$w.file.list index current] end
		if ![string length $filename] {
			$w.file.list mark set current 1.0
			tag_line $w.file.list
		}
#		puts [$w.file.list index current]
		
		bind $w.file.list <1> { tag_line .wsel.file.list }
#		bind $w.file.list <Double-1> { setSelectionAndDestroy }
#		bind $w.file.list <Double-1> {
#			tag_line .wsel.file.list
#			setSelectionAndDestroy
#		}
	
		lower . .wsel
		tkwait window $w
	}
#	puts "$oldfilename --- $filename"
#	puts "$filename ($imgtype) $project_item"
	if { [string compare $oldfilename $filename] != 0 } {
		loadImage $filename -1
	}
	updateParametersAndImage
}

proc setSelectionAndDestroy { } {
	global filename
	global imgtype
	global project_item
	set item [.wsel.file.list get -- seline.first seline.last]
	set list [split $item ":"]
	set filename [lindex $list 0]
	set imgtype [lindex $list 1]
	set project_item "[lindex $list 2]:[lindex $list 3]"
#	puts "$filename ($imgtype) $project_item"
	destroy .wsel
}

proc updateClassParam { } {
	global project_item theimg
	if [regexp "Class" $project_item] {
		set n [Bimage get $theimg nimages]
		for { set i 0 } { $i < $n } { incr i } {
			Bimage set $theimg select $i [Bmg get $project_item select [expr $i + 1]]
		}
	}
}

proc updateParametersAndImage { } {
	global filename paramfile modelfile
	global theimg imgtype
	global project_item mg_sel
	
#	puts "Updating parameters and image: $project_item"
	
	if { [Bmg exists] } {
		Bmg set all active 0
		if [regexp "Reconstruction" $project_item] { Bmg set all active 1 }
		if { $imgtype == "frame" } {
			if { [Bimage get $theimg nslices] > 1 } { switchImage }
		}
		set mg_sel [Bmg get $project_item select]
		wm title . [list $filename ":" $paramfile]
#		puts [Bmg get all number_of_rec]
	} elseif { [Bmodel exists] } {
		wm title . [list $filename ":" $modelfile]
	}
	
#	Bimage set $theimg label $project_item
	updatePixelsize
	updateOrigins

	if { [Bmg exists] } {
		getCTFparam
		updateClassParam
		updateBoxParam
		updateNodeParam
		updateTomoTable
		updateMarkerParam
		updateSpotParam
		updateLayerLineParam
#		puts [Bmg get all number_of_rec]
	} elseif { [Bmodel exists] } {
		objectsDrawAll
	}
}

proc getNeighborImageOfType { dir } {
	global filename
	global theimg imgtype
	global project_item

	set wc [getControlWindow $theimg]
	set img_num [expr [$wc.image.scale get] + $dir]
	set img_max [$wc.image.scale cget -to]
	
	if { $img_num >= 0 && $img_num <= $img_max } {
		$wc.image.scale set $img_num
		updateParametersAndImage
		return
	}
	
	if { [Bmg exists] } {
		set img_files [Bmg get all image_filenames]
	} elseif { [Bmodel exists] } {
		set img_files [Bmodel get all image_filenames]
	} else {
		return
	}
	
#	set oldfilename $filename
	set f ""
	set prevfile ""
	set prevlist ""
	set nxt 0
	foreach name $img_files {
		set list [split $name ":"]
		if [string equal $imgtype [lindex $list 1]] {
			set f [lindex $list 0]
			if { $nxt } {
				break
			}
			if [string equal $filename $f] {
				if { $dir > 0 } {
					set nxt 1
					set f ""
				} else {
#					if [string length $prevfile] { set f $prevfile }
#					if [string length $prevlist] { set list $prevlist }
					set f $prevfile
					set list $prevlist
					break
				}
			}
			set prevfile $f
			set prevlist $list
		}
	}

	if [string length $f] {
		set filename $f
		set project_item "[lindex $list 2]:[lindex $list 3]"
		loadImage $filename -1
		updateParametersAndImage
	}
}

proc getNeighborImageOfType2 { dir } {
	global filename
	global theimg imgtype
	global project_item

	set wc [getControlWindow $theimg]
	set img_num [expr [$wc.image.scale get] + $dir]
	
	if { [Bmg exists] } {
		set img_files [Bmg get all image_filenames]
	} elseif { [Bmodel exists] } {
		set img_files [Bmodel get all image_filenames]
	} else {
		return
	}
	
#	set oldfilename $filename
	set f ""
	set curr_item ""
	set prevfile ""
	set prevlist ""
  	set nxt 0
#	puts $project_item
	foreach name $img_files {
		set list [split $name ":"]
#		puts $name
		if [string equal $imgtype [lindex $list 1]] {
			set f [lindex $list 0]
			set curr_item "[lindex $list 2]:[lindex $list 3]"
			if { $nxt } {
				break
			}
#			puts $curr_item
			if [string equal $project_item $curr_item] {
#				puts $curr_item
				if { $dir > 0 } {
					set nxt 1
				} else {
					if [string length $prevfile] { set f $prevfile }
					if [string length $prevlist] { set list $prevlist }
					break
				}
			}
			set prevfile $f
			set prevlist $list
		}
	}

	if [string length $f] {
		set project_item "[lindex $list 2]:[lindex $list 3]"
		if [string equal $filename $f] {
		} else {
			set filename $f
			loadImage $filename -1
		}
		$wc.image.scale set $img_num
		updateParametersAndImage
	}
}

## @brief Sets up micrograph id and total.
#
# @param	w				window.

proc setupMicrographID { w } {
	global project_item
	
	set idx [Bmg get $project_item index_of_mg]
	set nmg [Bmg get all number_of_mg]

	if ![winfo exists $w] {
		frame $w
	}
	
	label $w.tag -text "Micrograph" -width 10 -anchor e
	label $w.str -width 50 -text $project_item -relief sunken -anchor w
	label $w.idx -width 5 -text $idx -anchor w
	label $w.nmg -width 5 -text " / $nmg" -anchor w
	
	pack $w.tag $w.str $w.idx $w.nmg -side left -padx 1m
}

## @brief Sets up micrograph navigation and selection buttons.
#
# @param	w				window.
# @param	typelabel		type of navigation (micrograph).

proc setupMicrographButtons { w typelabel } {
	global mg_sel
	
	if ![winfo exists $w] {
		frame $w
	}
	
	button $w.prev -text "Previous $typelabel" \
				-relief raised -command "getNeighborImageOfType -1"
	button $w.next -text "Next $typelabel" \
				-relief raised -command "getNeighborImageOfType 1"
	checkbutton $w.select -text "Select" -variable mg_sel \
				-command { selectMicrograph }
	pack $w.prev $w.next $w.select \
			-side left -pady 2 -padx 2 -ipadx 2 -ipady 2
}

## @brief Creates micrograph parameters from the image.
#

proc createMgParam { } {
	global filename theimg imgtype
	global paramfile project_item
	if [Bmg exists] { return }
	set id_name [file rootname [file tail $filename]]
#	puts "New ID: $id_name"
	set nz [Bimage get $theimg nslices]
	if { $imgtype == "" } { setImageType }
	if { $imgtype == "mg" || $imgtype == "frame" } {
		if { $nz > 1 } { switchImage }
		set nz [Bimage get $theimg nslices]
	}
	set nimg [Bimage get $theimg nimages]
	Bmg create $filename $imgtype
	if { $nz > 1 } {
		Bmg set all active 1
		if { $nimg > 1 } {
			set imgtype "part"
		} else {
			set imgtype "rec"
		}
		set project_item "Reconstruction:$id_name"
		Bmg set $project_item id $id_name
	} else {
		Bmg set all active 0
		set project_item "Micrograph:$id_name"
		Bmg set $project_item id $id_name
		if { $nimg > 1 && $imgtype != "frame" } {
			set project_item "Field:$id_name"
			Bmg set $project_item id $id_name
		}
	}
#	puts "New item: $project_item"
	set origin [Bimage get $theimg origin]
	set ps [Bimage get $theimg pixel_size]
	Bmg set $project_item origin [lindex $origin 0] [lindex $origin 1] [lindex $origin 2]
	Bmg set $project_item pixel_size $ps
#	puts "Pixel size: $ps"
#	Bimage set $theimg label $project_item
}

## @brief Returns the current project item.
#
# @param	n			micrograph index (first = 0)

proc micrographItem { n } {
	global project_item

	if { ![Bmg exists] && ![Bmodel exists] } { return "" }
	
	if [regexp "Model" $project_item] {
		return $project_item
	} elseif [regexp "Reconstruction" $project_item] {
		return $project_item
	} elseif [regexp "Class" $project_item] {
		return $project_item
	} elseif [regexp "Micrograph" $project_item] {
		set id [Bmg get $project_item id $n]
		return "Micrograph:$id"
#		return $project_item
	} elseif [regexp "Field" $project_item] {
		set id [Bmg get $project_item id $n]
		return "Micrograph:$id"
	} else {
		puts "Error: project_item not defined! ($project_item)"
	}
}

proc selectMicrograph { } {
	global project_item mg_sel
	Bmg set $project_item select $mg_sel
}

## @brief Updates an existing project for micrograph parameters in memory.
#

proc setMgParam { } {
	global filament_width filament_node_radius
	global marker_radius fom_cutoff
	global theimg imgtype
	global filename project_item
	if ![Bmg exists] { createMgParam }
	puts "setMgParam: $project_item"
#	Bimage set $theimg label $project_item
	if { [Bimage get $theimg nslices] > 1 } {	# Converting from mg to rec
		Bmg set all active 1
		if { [string first "mg" $imgtype] > -1 } {
			set imgtype "rec"
			puts "Trying to convert to reconstruction"
		}
#		puts "Reconstruction: $imgtype $filename"
	} else {
		if { [string first "rec" $imgtype] > -1 } { set imgtype "mg" }
	}
#	puts "Bmg set $project_item filename $filename $imgtype"
	Bmg set $project_item filename $filename $imgtype
	Bmg set $project_item pixel_size [.pixel_size.entry get]
#	if { [Bimage get $theimg nslices] < 2 } {	# Micrograph parameters only
#		if [winfo exists .wctf] { setCTFparam }
#	}
	puts "setMgParam: $filename"
	setBoxSize
	if [winfo exists .wfil] {
		set filament_width [.wfil.fil_width.entry get]
		set filament_node_radius [.wfil.node_radius.entry get]
		Bmg set $project_item filament_width $filament_width
		Bmg set $project_item filament_node_radius $filament_node_radius
	}
	if [winfo exists .wtomo] {
		set marker_radius [.wtomo.marker_radius.e get]
		Bmg set $project_item marker_radius $marker_radius
	}
}

## @brief Saves the micrograph data base to a parameter file.
#

proc writeParam { } {
	global filename
	global paramfile paramdir
	set currdir $paramdir
	if { [string length $currdir] < 2 } { set currdir [pwd] }
	if { [string first $currdir ".."] > -1 } { set currdir [pwd] }
	puts "Current directory: $currdir"
	createMgParam
	set thename [file rootname [file tail $filename]]
	if { [string length $paramfile] < 1 } {
		set paramfile $thename
		if { ![string match "*.star" $paramfile] } {
			append paramfile ".star"
		}
	}
	set paramfile [file tail $paramfile]
    set types {
		{"STAR files"		{.star .STAR}	}
		{"XML files"		{.xml .XML}	}
		{"EMX files"		{.emx .EMX}	}
	}
	set paramfile [tk_getSaveFile -filetypes $types \
			-initialfile $paramfile -initialdir $currdir -defaultextension .star]
	if { [string length $paramfile] < 1 } { return }
	set paramfile [relativePath $paramdir $paramfile]
	setMgParam
	wm title . [list $filename ":" $paramfile]
#	puts "Writing $paramfile"
	Bmg write $paramfile
#	puts "Writing .bshow"
	writeSettingsFile
}

## @brief Edits a text parameter file.
#

proc editParam { } {
	global paramfile paramdir
	editTextFile $paramfile
}

## @brief Checks if the parameter file exist, otherwise save it.
#

proc checkParamFile { } {
	global paramfile
	if { ![file exists $paramfile] } {
		writeParam
	}
}


